/*
   Find out best unrolling steps and generate the final Fortran code.

   This program calls 'search' for every set of unrolling steps,
   described in a parameter file supplied as command-line argument
   for 'gen'.

   By convention, the parameter file names have .gen as extension.
*/

#include <stdio.h>
#include <stdlib.h>

#define MAX_NBR		100	/* maxi. # of sets for unrolling steps */

static FILE *f_par, *f_dat;
static char line[1024], progname[1024], datafile[1024], s[1024];

int nbr=0;
struct {
    int dim;
    char prefix[10];
    int *best;		/* to store best unrolling steps */
} record[MAX_NBR];

static int mode=0;

int main(int argc, char *argv[])
{
    char *p, *p1=NULL;
    int i, j, step, lineno=0;

    for (i=1;i<argc;i++) {
      p=argv[i];
      if (*p=='-') {
        if (!*(++p)) continue;
        switch (toupper(*p)) {
	  case 'N' :   /* don't run search */
	    mode=1;
	    break;
#if 0
	  case 'D' :   /* defines for m4 */
	    sprintf(defs+strlen(defs), " %s", p);
	    break;
#endif
	  case 'T' :   /* test with arbitrairily assigned unrolling steps */
	    mode=2;
	    break;
          default  :
            fprintf(stderr,"Invalid option \"%s\".\n",argv[i]);
      usage:
            fprintf(stderr,"Usage: gen [options] gen-file [options]\n");
            fprintf(stderr,"Options:\n");
            fprintf(stderr,"\t-n\tDon't run \"search\", use existing *.search.* files.\n");
            fprintf(stderr,"\t-t\tGenerate .f file with arbitrarily assigned unrolling steps.\n");
            return 1;
        }
      } else {
          if (p1!=NULL) {
            printf("Too many file names!\n"); goto usage;
          }
          p1=argv[i];
      }
    }

    if (p1==NULL) goto usage;

    if ((f_par=fopen(p1, "rt"))==NULL) {
	fprintf(stderr,"Cannot open parameter file \"%s\"!\n", p1);
	goto usage;
    }

    line[1023]='\0';
    fgets(line,1023,f_par); sscanf(line,"%s",progname);
    lineno++;
    do {
        if (f_par==NULL) break;
	if (fgets(line, 1023, f_par)==NULL) break;
	lineno++;
	/* cancel leading spaces */
	for (p=line; isspace(*p); p++);
	if (*p=='\0') continue;		/* ignore blanc line */
	if (nbr>=MAX_NBR) {
	    fprintf(stderr,"Too many parameter sets!\n");
	    fclose(f_par);
	    exit(1);
	}
	if (sscanf(p,"%d %d %s", &(record[nbr].dim), &step,
                   record[nbr].prefix)!=3) {
	    fprintf(stderr,"Invalid input at line %d!\n",lineno);
	    fclose(f_par);
	    exit(1);
        }
	if ((record[nbr].best=malloc(record[nbr].dim*sizeof(int)))==NULL) {
	    fprintf(stderr,"Memory allocation error!\n");
	    fclose(f_par);
	    exit(1);
	}
	if (!mode) {
            /* construct data file for "search" */
	    sprintf(datafile,"%s.data.%s", progname, record[nbr].prefix);
            if ((f_dat=fopen(datafile,"w+t"))==NULL) {
                fprintf(stderr,"Cannot create file \"%s\"!\n", datafile);
                fclose(f_par);
                exit(1);
            }
        }
        do {
            if (fgets(line,1023,f_par)==NULL) {
                fclose(f_par); f_par=NULL;
                break;
            }
            lineno++;
            for (p=line; isspace(*p); p++);
            if (*p=='\0') break;
            if (!mode) fputs(p,f_dat);
        } while (1);
        if (!mode) {
            fclose(f_dat);
	    /* execute "search" to determine best unrolling steps */
	    sprintf(line,"search %d %s %s %d %s",
	                 record[nbr].dim,
	                 progname, datafile, step, record[nbr].prefix);
	    printf("Executing \"%s\"\n",line);
            if (system(line)) {
                fprintf(stderr, "Cannot execute search!\n");
                if (f_par!=NULL) fclose(f_par);
                exit(1);
	    }
	}
	if (mode<=1) {
	    /* read best unrolling steps from the output file */
	    sprintf(datafile,"%s.search.%s",progname,record[nbr].prefix);
	    if ((f_dat=fopen(datafile,"rt"))==NULL) {
	        fprintf(stderr,"Cannot open \"%s\"!\n",datafile);
                if (f_par!=NULL) fclose(f_par);
                exit(1);
	    }
	    line[0]=s[0]='\0';
	    do {
	        if (fgets(s,1023,f_dat)==NULL) break;
                for (p=s; isspace(*p); p++);
                if (*p=='\0') continue;
                strcpy(line,p);
	    } while (1);
	    fclose(f_dat);
	    /* extract the unrolling steps */
	    strcpy(s,"%*s %*s %*s");
	    for (i=0; i<record[nbr].dim; i++) {
	        strcat(s," %d");
	        if (sscanf(line, s, record[nbr].best+i)!=1) {
		    fprintf(stderr,"Error in reading \"%s\".",datafile);
		    if (f_par!=NULL) fclose(f_par);
		    exit(1);
	        }
	        strcpy(s+strlen(s)-2,"%*s");
	    }
            if (!mode) {
		printf("Unrolling steps = ");
		for (i=0; i<record[nbr].dim; i++)
		    printf("%d ",record[nbr].best[i]);
		printf("\n");
	    }
	} else {
	    for (i=0; i<record[nbr].dim; i++) record[nbr].best[i]=i+3;
	}
	nbr++;
    } while (1);
    if (f_par!=NULL) fclose(f_par);

    printf("Creating %s.f with:\n",progname);
    strcpy(line, "m4 -P -D_RELEASE_");
    for (j=0; j<nbr; j++) {
	for (i=0; i<record[j].dim; i++) {
	    sprintf(line+strlen(line), " -D%s%d=%d",
	        record[j].prefix, i+1, record[j].best[i]);
	    printf("\t%s%d=%d\t", record[j].prefix, i+1, record[j].best[i]);
	}
	printf("\n");
    }
    sprintf(line+strlen(line), " defs.m4 %s.m4 | m4post | do-enddo -u >%s.f",
            progname, progname);
    if (system(line)) {
        fprintf(stderr, "Cannot execute \"%s\"!\n", line);
    }

    exit(0);
}
